%% 2D COMPLIANCE-CONSTRAINED VOLUME-MINIMIZATION TOPOLOGY OPTIMIZATION %%
%%%%%%%%%%%%%%% using direct Lagrange multiplier method %%%%%%%%%%%%%%%
%%%%%%%%%%%%%% with a linear approximation to constraint %%%%%%%%%%%%%%
% INPUTS:
% nelx          :      number of elements in x-direction
% nely          :      number of elements in y-direction
% penal         :      penalty for SIMP interpolation
% rmin          :      Filter radius
% ft            :      Filter type (1 = sensitivity, 2 = density)
% compconst     :      compliance constraint
%
% Try running   :      minV_LinDirect(200,100,3,2,2,136.1)
function minV_LinDirect(nelx,nely,penal,rmin,ft,compconst)
%% MATERIAL PROPERTIES
E0 = 1;
Emin = 1e-9;
nu = 0.3;
%% PREPARE FINITE ELEMENT ANALYSIS
A11 = [12  3 -6 -3;  3 12  3  0; -6  3 12 -3; -3  0 -3 12];
A12 = [-6 -3  0  3; -3 -6 -3 -6;  0 -3 -6  3;  3 -6  3 -6];
B11 = [-4  3 -2  9;  3 -4 -9  4; -2 -9 -4 -3;  9  4 -3 -4];
B12 = [ 2 -3  4 -9; -3  2  9 -2;  4  9  2  3; -9 -2  3  2];
KE = 1/(1-nu^2)/24*([A11 A12;A12' A11]+nu*[B11 B12;B12' B11]);
nodenrs = reshape(1:(1+nelx)*(1+nely),1+nely,1+nelx);
edofVec = reshape(2*nodenrs(1:end-1,1:end-1)+1,nelx*nely,1);
edofMat = repmat(edofVec,1,8)+repmat([0 1 2*nely+[2 3 0 1] -2 -1],nelx*nely,1);
iK = reshape(kron(edofMat,ones(8,1))',64*nelx*nely,1);
jK = reshape(kron(edofMat,ones(1,8))',64*nelx*nely,1);
% Define loads and supports (cantilever with two point loads)
F = sparse(2*(nelx+1)*(nely+1)-1:2*(nelx+1)*(nely+1),1,[-0.5 -1],2*(nely+1)*(nelx+1),1);
fixeddofs = 1:2*(nely+1);
U = zeros(2*(nely+1)*(nelx+1),1);
alldofs = 1:2*(nely+1)*(nelx+1);
freedofs = setdiff(alldofs,fixeddofs);
%% PREPARE FILTER
iH = ones(nelx*nely*(2*(ceil(rmin)-1)+1)^2,1);
jH = ones(size(iH));
sH = zeros(size(iH));
k = 0;
for i1 = 1:nelx
    for j1 = 1:nely
        e1 = (i1-1)*nely+j1;
        for i2 = max(i1-(ceil(rmin)-1),1):min(i1+(ceil(rmin)-1),nelx)
            for j2 = max(j1-(ceil(rmin)-1),1):min(j1+(ceil(rmin)-1),nely)
                e2 = (i2-1)*nely+j2;
                k = k+1;
                iH(k) = e1;
                jH(k) = e2;
                sH(k) = max(0,rmin-sqrt((i1-i2)^2+(j1-j2)^2));
            end
        end
    end
end
H = sparse(iH,jH,sH);
Hs = sum(H,2);
%% INITIALIZE ITERATION
x = ones(nely,nelx);
xPhys = x;
loop = 0;
change = 1;
%% START ITERATION
while (change > 1e-3 && loop < 200)
    loop = loop + 1;
    %% FE-ANALYSIS
    sK = reshape(KE(:)*(Emin+xPhys(:)'.^penal*(E0-Emin)),64*nelx*nely,1);
    K = sparse(iK,jK,sK); K = (K+K')/2;
    U(freedofs) = K(freedofs,freedofs)\F(freedofs);
    %% OBJECTIVE FUNCTION AND SENSITIVITY ANALYSIS
    ce = reshape(sum((U(edofMat)*KE).*U(edofMat),2),nely,nelx);
    comp = sum(sum((Emin+xPhys.^penal*(E0-Emin)).*ce));
    dc = -penal*(E0-Emin)*xPhys.^(penal-1).*ce;
    dv = ones(nely,nelx)/nelx/nely;
    if (loop == 1 && nargin < 7)
        compconst = 2*comp;
    end
    f0val = mean(xPhys(:));
    fval = comp/compconst - 1;
    %% FILTERING/MODIFICATION OF SENSITIVITIES
    if ft == 1
        dc(:) = H*(x(:).*dc(:))./Hs./max(1e-3,x(:));
    elseif ft == 2
        dc(:) = H*(dc(:)./Hs);
        dv(:) = H*(dv(:)./Hs);
    end
    %% OPTIMALITY CRITERIA UPDATE OF DESIGN VARIABLES AND PHYSICAL DENSITIES
    change1 = true;eta = 0.5;ocIter = 0;move = 0.2;
    dc = dc(:);x1 = x(:);dcx1 = dc.*x1;
    R = ((-dc./dv(:)).^eta);gRem = (compconst - comp);
    xMax = min(x1+move,1);xMin = max(x1-move,1e-10);
    varIn = true(nelx*nely,1);
    gToDist = gRem+sum(dcx1);
    cUps = dc.*(xMax-x1); cDns = dc.*(xMin-x1);
    while(change1)
        xnew = x1.*R*gToDist/ sum(dcx1.*R.*varIn);
        upLgc=xnew>xMax;dnLgc=xnew<xMin;
        change1 = ~isequal(~(upLgc|dnLgc),varIn);
        varIn = ~(upLgc|dnLgc);ocIter = ocIter + 1;
        gToDist=gRem -sum(cUps(upLgc)) -sum(cDns(dnLgc)) + sum(dcx1(varIn));
        if (gToDist>=0||ocIter>15)
            if move<1
                move = min(move*2,1);ocIter = 0;
                xMax = min(x1+move,1);xMin = max(x1-move,1e-10);
                gToDist = gRem+sum(dcx1);varIn = true(nelx*nely,1);
                cUps = dc.*(xMax-x1); cDns = dc.*(xMin-x1);
            else
                break;  
            end
        end
    end
    xnew(xnew>xMax) = xMax(xnew>xMax);xnew(xnew<xMin) = xMin(xnew<xMin);
    %end
    change = full(max(abs(xnew(:)-x(:))));
    x = xnew;
    if ft == 1
        xPhys = xnew;
    elseif ft == 2
        xPhys(:) = (H*xnew(:))./Hs;
    end
    %% PRINT RESULTS
    fprintf('It.:%5i Con.:%11.3e Vol.:%11.3e ch.:%7.3f\n',...
        loop,fval,f0val,change);
    %% PLOT DENSITIES
    colormap(jet); imagesc(1-xPhys); caxis([0 1]); axis equal; axis off; drawnow;
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This code was written by Oded Amir,                                      %
% Faculty of Civil & Environmental Engineering,                            %  
% Technion - Israel Institute of Technology.                               %
%                                                                          %
% Contact: odedamir@technion.ac.il                                         %
% Modified for direct Lagrange multiplier update by Tej Kumar and Krishnan %                                                                         
% Suresh, Department of Mechanical Engineering, UW Madison.                %
%                                                                          %
% The code can be  downloaded from the web-site:                           %
% http://www.ersl.wisc.edu/software/DirectLagrangeMultiplier.zip           %
%                                                                          %   
% If using this code please cite: T. Kumar and K. Suresh, "Direct Lagrange % 
% Multiplier Updates in Topology Optimization Revisited" Struct Multidisc  %                                                                  
% Optim,  (Accepted)                                                       %
%                                                                          %
% Details are discussed in the paper:                                      %
% "Revisiting approximate reanalysis in topology optimization: on the      %
% advantages of recycled preconditioning in a minimum weight procedure",   %
% http://link.springer.com/article/10.1007/s00158-014-1098-7               %
%                                                                          %
% Disclaimer:                                                              %
% The author reserves all rights but does not guarantee that the code is   %
% free from errors. Furthermore, the author shall not be liable in any     %
% event caused by the use of the program.                                  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
