%% MULTIRESOLUTION VOLUME-CONSTRAINED COMPLIANCE-MINIMIZATION TOPOLOGY OPTIMIZATION %%
%%%%%%%%%%%%%%% using direct Lagrange multiplier method %%%%%%%%%%%%%%%
% INPUTS:
% filename      :      Input file containing mesh and boundary conditions
% nDenElDim     :      Number of density element per element per dimension
% nDesElDim     :      Number of design element per element per dimension
% volFrac       :      Volume fraction constraint
% p             :      Penalty for SIMP interpolation
% filterRad     :      Filter radius
% ft            :      Filter type (1 = sensitivity, 2 = density)
% ocType        :      optimality criteria using 1=bisection, 2=direct
%
% Try running   :      multiResolution('elem_L1600.dat',6,6,0.5,3,1.5,2,2)
function multiResolution(fileName,nDenElDim,nDesElDim,volFrac,p,filterRad,ft,ocType)
dim = 2;    t = 1;      epsX = 0.001;
[conn,nodeRaw,matProp,dispBC,forceBC,nNode,nElem,dof] = readData(fileName,dim);
%Modifying node data to handle missing node numbers
max_nodeNo=max(nodeRaw(:,1));       node=zeros(max_nodeNo,3);
node(nodeRaw(:,1),1)=1:length(nodeRaw(:,1));
node(nodeRaw(:,1),2)=nodeRaw(:,2);  node(nodeRaw(:,1),3)=nodeRaw(:,3);
%% Displacement elements
x = node(:,2);                  y = node(:,3);
xmax = max(x);                  ymax = max(y);
xs = unique(x);                 ys = unique(y);
nelx = length(xs)-1;            nely = length(ys)-1;
dx = xmax/nelx;                 dy = ymax/nely;
nDofs = dof*nNode;              nn = node(:,1);
%Diplacement element dofs of all elements
eDofs=[2*node(conn(:,3),1)-1  2*node(conn(:,3),1) ...
    2*node(conn(:,4),1)-1 2*node(conn(:,4),1) ...
    2*node(conn(:,5),1)-1 2*node(conn(:,5),1) ...
    2*node(conn(:,6),1)-1 2*node(conn(:,6),1) ];
iK = reshape(kron(eDofs,ones(8,1))',64*nElem,1);
jK = reshape(kron(eDofs,ones(1,8))',64*nElem,1);
%Displacement element centers
dispCen=[mean(reshape(node(conn(:,3:6)',2),4,nElem),1)' ...
    mean(reshape(node(conn(:,3:6)',3),4,nElem),1)'];
%% Add density elements
nDenEl = nDenElDim^dim;         areaPerDenEl = dx*dy/nDenEl;
%Density element centers
denCenX = reshape(kron(linspace(dx/(2*nDenElDim),dx-dx/(2*nDenElDim),nDenElDim)'-dx/2,ones(nDenElDim,1)),[],1);
denCenY = reshape(kron(ones(1,nDenElDim),linspace(dy/(2*nDenElDim),dy-dy/(2*nDenElDim),nDenElDim)-dy/2),[],1);
%Density element centers for entire domain i.e. for all displacement elems
denCenXAll = (repmat(dispCen(:,1),1,nDenEl)+repmat(denCenX',nElem,1))';
denCenYAll = (repmat(dispCen(:,2),1,nDenEl)+repmat(denCenY',nElem,1))';
%Initialize variables to store data during optimization
denVar = zeros(nDenEl,nElem);       gradDen = zeros(nDenEl,nElem);
%% Add design elements
nDesEl = nDesElDim^dim;         arDesEl = dx*dy/nDesEl;
%Design element centers
desCenX = reshape(kron(linspace(dx/(2*nDesElDim),dx-dx/(2*nDesElDim),nDesElDim)'-dx/2,ones(nDesElDim,1)),[],1);
desCenY = reshape(kron(ones(1,nDesElDim),linspace(dy/(2*nDesElDim),dy-dy/(2*nDesElDim),nDesElDim)-dy/2),[],1);
%Design element centers for entire domain i.e. for all displacement elems
desCenXAll = (repmat(dispCen(:,1),1,nDesEl)+repmat(desCenX',nElem,1))';
desCenYAll = (repmat(dispCen(:,2),1,nDesEl)+repmat(desCenY',nElem,1))';
%Initialize variables to store data during optimization
desVar = volFrac*ones(nDesEl,nElem);        gradDes = zeros(nDesEl,nElem);
%% Get base material elasticity matrix and stiffness templates
YM = matProp(1,2);  nu = matProp(1,3);
D = getElasticityTensor(YM,nu);
%Compute elemental stiffness matrix for base material at Gpts
gptX = 2*denCenX/dx;  gptY = 2*denCenY/dy;  gpts = [gptX,gptY];
[B,Det] = calc_Bdet1(conn,node,gpts);
[KeGpt] = Kelem_common(D,B,Det,(4/size(B,2))*ones(1,size(B,2)),t);
KeGptRshpd = reshape(KeGpt,[],size(gpts,1));
%% Initializing displacement and force vectors
u = zeros(nDofs,1);
f = sparse((nn(forceBC(:,1))-1)*dof+forceBC(:,2),1,forceBC(:,3),nDofs,1);
[prscDofs,unique_ID] = unique((nn(dispBC(:,1))-1)*dof+dispBC(:,2));
u(prscDofs) = dispBC(unique_ID,3);
freeDofs=setdiff(1:nDofs,prscDofs);
%% Defining projection to compute density variables
rad = filterRad*(dx+dy)/2;
allDispElem = (1:nElem)';
iV = zeros(ceil((2*filterRad)^2*nDesEl*nDenEl*nElem),1);
jV = zeros(ceil((2*filterRad)^2*nDesEl*nDenEl*nElem),1);
kV = zeros(ceil((2*filterRad)^2*nDesEl*nDenEl*nElem),1);
ctr = 0;
for el = 1:nElem
    thisElsNeibh = and(and(dispCen(:,1)>(min(dispCen(el,1))-(rad+dx)),...
        dispCen(:,1)<(max(dispCen(el,1))+(rad+dx))),...
        and(dispCen(:,2)>(min(dispCen(el,2))-(rad+dy)),...
        dispCen(:,2)<(max(dispCen(el,2))+(rad+dy))));
    dist1 = ((reshape(denCenXAll(:,el),[],1)-reshape(desCenXAll(:,thisElsNeibh),1,[])).^2 ...
        +(reshape(denCenYAll(:,el),[],1)-reshape(desCenYAll(:,thisElsNeibh),1,[])).^2);
    ins = lt(dist1,rad^2);
    is = repmat(reshape((1:nDenEl)'+(el-1)*nDenEl,[],1),1,sum(thisElsNeibh)*nDesEl);
    js = repmat(reshape((1:nDesEl)'+(allDispElem(thisElsNeibh)'-1)*nDesEl,1,[]),nDenEl*length(el),1);
    kIn = rad-sqrt(dist1(ins(:)));
    ctrNew = ctr+length(kIn);
    iV(ctr+1:ctrNew) = is(ins);
    jV(ctr+1:ctrNew) = js(ins);
    kV(ctr+1:ctrNew) = kIn;
    ctr = ctrNew;
end
H = sparse(iV(1:ctr),jV(1:ctr),kV(1:ctr),nDenEl*nElem,nDesEl*nElem);%sparse(iIn,jIn,kIn,denResPerElem*nElem,desResPerElem*nElem);
clear iV jV kV dist1 is js kIn
Hs = full(sum(H,2));
%% Start iterations
%Some optimization parameters before iteration
noOfIterations = 500;               totVol = volFrac*nElem*dx*dy;
obj = zeros(noOfIterations,1);      change = zeros(noOfIterations,1);
for iter = 1:noOfIterations
    %Continuation scheme
    if rem(iter,20) == 0
        p = min(p+0.5,4);
    end
    %Compute density from design variables
    denVar(:) = (H*desVar(:))./Hs;
    if iter == 1, denVec = denVar(:); end
    %Compute Ke using density variables
    Ke = squeeze(sum(KeGptRshpd.*reshape(denVar.^p,1,nDenEl,nElem),2));
    %Compute K stiffness matrix
    K = sparse(iK,jK,Ke(:),nDofs,nDofs);
    %Compute u
    u(freeDofs) = K(freeDofs,freeDofs)\f(freeDofs);
    %Compute objective
    obj(iter) = f'*u;
    if iter == 1
        scaling_Factor = obj(iter);%1;%
        disp(['Scaling objective by factor = ' num2str(scaling_Factor)]);
    end
    %Compute gradient wrt density variable
    for i = 1:nDenEl
        gradDen(i,:) = -p*sum((u(eDofs)*KeGpt(:,:,i)).*u(eDofs),2)'.*denVar(i,:).^(p-1);
    end
    %Compute gradient wrt design variables
    gradDes(:) = H'*(gradDen(:)./Hs);
    %Apply scaling
    obj(iter) = obj(iter)/scaling_Factor;
    gradDes = gradDes/scaling_Factor;
    %Update design variables
    desVec = desVar(:);dc = gradDes(:);dv = arDesEl*ones(size(desVar(:)));   
    switch ocType
        case 1 %Bisection from 88-line code
            l1 = 0; l2 = 1e9; move = 0.2;ocs = 0;
            while (l2-l1)/(l1+l2) > 1e-3
                lmid = 0.5*(l2+l1);
                desVecNew = max(epsX,max(desVec-move,min(1,min(desVec+move,desVec.*(-dc./dv/lmid).^0.5))));
                if ft == 1
                    denVec = desVecNew;
                    if sum(denVec(:))*arDesEl > totVol, l1 = lmid; else, l2 = lmid; end
                elseif ft == 2
                    denVec(:) = (H*desVecNew(:))./Hs;
                    if sum(denVec(:))*areaPerDenEl > totVol, l1 = lmid; else, l2 = lmid; end
                end
                ocs = ocs + 1;
            end
        case 2 %Direct method
            move = 0.2;eta = 0.5;ocs = 0;
            ins = true(numel(desVec),1);
            dMax = min(desVec+move,1);dMin = max(desVec-move,epsX);
            newDesMult = desVec.*((-dc./dv).^eta);
            gRem = totVol/arDesEl; gToDist = gRem/sum(newDesMult);
            if ft == 1
                while true
                    desVecNew = newDesMult*gToDist;
                    ups = desVecNew>dMax;dns = desVecNew<dMin;ocs = ocs + 1;
                    ins1 = ~(ups|dns);
                    gToDist = (gRem-sum(dMax(ups))-sum(dMin(dns)))/sum(newDesMult(ins1));
                    if isequal(ins,ins1), break; end
                    ins = ins1;
                end
            elseif ft == 2
                temp1 = zeros(numel(desVec),1);    temp2 = zeros(numel(desVec),1);
                change1 = true;gToDist = gRem;
                while(change1)
                    temp1(:) = 0;temp1(ins) = newDesMult(ins);
                    desVecNew = newDesMult/(sum((H*temp1)./Hs)/gToDist);
                    ups = desVecNew>dMax;dns = desVecNew<dMin;
                    temp2(:) = 0;temp2(ups) = dMax(ups);temp2(dns) = dMin(dns);
                    gToDist = gRem - sum((H*(temp2))./Hs);
                    change1 = ~isequal(~ins,(ups|dns));
                    ins = ~(ups|dns);ocs = ocs + 1;
                end
            end
            desVecNew(ups) = dMax(ups);desVecNew(dns) = dMin(dns);
    end
    change(iter) = max(abs(desVecNew(:)-desVec(:)));
    desVar = desVecNew;
    %Display values
    fprintf('Iter. %d, obj. = %8.5f, ch. = %5.4f, vol.(des,den) = (%5.4f,%5.4f), oc = %d\n',iter,obj(iter),change(iter),sum(desVar,'all')*arDesEl,sum(denVar,'all')*areaPerDenEl,ocs);
    %Check for convergence
    if(change(iter)<0.01)
        disp('Change in density converged');
        break;
    end
end
%printing the final design
vxDen = (reshape(denCenXAll,[],1)+[-dx/nDenElDim/2 dx/nDenElDim/2 +dx/nDenElDim/2 -dx/nDenElDim/2])';
vyDen = (reshape(denCenYAll,[],1)+[-dy/nDenElDim/2 -dy/nDenElDim/2 +dy/nDenElDim/2 +dy/nDenElDim/2])';
figure;
patch(vxDen,vyDen,1-denVar(:),'EdgeColor','none');
colormap('gray');
axis equal off;
title('Density Variable Distribution');

vxDes = (reshape(desCenXAll,[],1)+[-dx/nDesElDim/2 dx/nDesElDim/2 +dx/nDesElDim/2 -dx/nDesElDim/2])';
vyDes = (reshape(desCenYAll,[],1)+[-dy/nDesElDim/2 -dy/nDesElDim/2 +dy/nDesElDim/2 +dy/nDesElDim/2])';
figure;
patch(vxDes,vyDes,1-desVar(:),'EdgeColor','none');
colormap('gray');
axis equal off;
title('Design Variable Distribution');
%-------------------------------------------------------------------------%
function [B,Det]=calc_Bdet1(Elem,Node,gpts)
nGpts = size(gpts,1);
B=cell(1,nGpts);    Det=cell(1,nGpts);
elem = 1;
XXElem =  Node(Elem(elem,3:6),2:3);

for gpt = 1:nGpts
    RI = gpts(gpt,1);    SI = gpts(gpt,2);
    [B1, Det1] = B_Det(XXElem,RI,SI);
    B{elem,gpt}=B1;    Det{elem,gpt}=Det1;
end
%-------------------------------------------------------------------------%
function [B0, Det] = B_Det(XX,R,S)
ZP = 1.0+R;     EP = 1.0+S;    ZM = 1.0-R;     EM = 1.0-S;
P=0.25*[-EM EM EP -EP;-ZM -ZP ZP ZM];   %dNdxi
XJ = P*XX;                              %dxdxi
Det = XJ(1,1)*XJ(2,2)-XJ(1,2)*XJ(2,1);  % Determinant
if(Det <= 0.0), disp('Jacobian is not positive definite'); return; end

XJI=[XJ(2,2) -XJ(1,2);-XJ(2,1) XJ(1,1)]/Det;    %inverse of Jacobian

R1=[1 0 0 0;0 0 0 1;0 1 1 0];
R2=zeros(4);    R2(1:2,1:2)=XJI;    R2(3:4,3:4)=XJI;
R3=zeros(4,8);  R3(1:2,1:2:7)=P;    R3(3:4,2:2:8)=P;
B0=R1*R2*R3;
%-------------------------------------------------------------------------%
function [Kgpt,K] = Kelem_common(D,B,Det,wt,t)
if nargin<5,    t = 1;  end
Kgpt=zeros(size(B{1,1},2),size(B{1,1},2),size(B,2));
for i=1:size(B,2)
    weight = wt(i)*t*Det{1,i};
    Kgpt(:,:,i) = B{1,i}'*D*B{1,i}*weight;
end
if nargout>1,    K = sum(Kgpt,3);   end
%-------------------------------------------------------------------------%
function [elemdata,nodedata,matprop,dispbc,forcebc,nNode,nElem,dof] = readData( fileName,dof )
fid = fopen(fileName,'r');
% Element data
fgets(fid); fgets(fid);
nElem = fscanf(fid,'%i\n',1);
fgets(fid); fgets(fid);
elemdata = fscanf(fid,'%i\n',[6,nElem])';
% Material property data
fgets(fid); fgets(fid);
NmaterialID = 1;
for i = 1:nElem
    if(NmaterialID < elemdata(i,2))
        NmaterialID = elemdata(i,2);
    end
end
matprop = fscanf(fid,'%g\n',[4,NmaterialID])';
% Nodal coordinates
fgets(fid);
nNode = fscanf(fid,'%i\n',1);
fgets(fid); fgets(fid);
nodedata = fscanf(fid,'%i %g %g\n',[dof+1, nNode])';
% Displacement boundary conditions
fgets(fid);
Ndispbc = fscanf(fid,'%i\n',1);
fgets(fid);
dispbc = fscanf(fid,'%i %g %g\n',[3, Ndispbc])';
% Force boundary conditions
fgets(fid);
Nforcebc = fscanf(fid,'%i\n',1);
fgets(fid);
forcebc = fscanf(fid,'%i %g %g\n',[3, Nforcebc])';
fclose(fid);
%-------------------------------------------------------------------------%
function D = getElasticityTensor(E,nu)
temp = E/(1-nu*nu);
D = temp*[1 nu 0;nu 1 0;0 0 (1-nu)/2];
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This Multiresolution Topology Optimization Matlab code is written by     %
% Tej Kumar, Department of Mechanical Engineering, UW-Madison              %
% Madison, Wisconsin, USA.                                                 %
%                                                                          %
% The code can be  downloaded from the web-site:                           %
% http://www.ersl.wisc.edu/software/DirectLagrangeMultiplier.zip           %
%                                                                          %  
% If using this code please cite: T. Kumar and K. Suresh, "Direct Lagrange % 
% Multiplier Updates in Topology Optimization Revisited" Struct Multidisc  %                                                                  
% Optim,  (Accepted)                                                       %
%                                                                          %
% Disclaimer:                                                              %
% The authors reserves all rights but do not guaranty that the code is     %
% free from errors. Furthermore, we shall not be liable in any event       %
% caused by the use of the program.                                        %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%